@@ -7,16 +7,15 @@
 # Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
 #
 
-from __future__ import absolute_import, division, unicode_literals
-
 import ast
 
-from mo_dots import is_data, is_null, Data, from_data
-from mo_future import text, number_types, binary_type
+from mo_dots import is_data, is_null, Data, from_data, exists, is_missing
+from mo_future import text, number_types, binary_type, is_text
+from mo_imports import expect
 from mo_parsing import *
 from mo_parsing.utils import is_number, listwrap
 
-unary_ops = None
+unary_ops = expect("unary_ops")
 
 
 class Call(object):
@@ -29,6 +28,7 @@ class Call(object):
 
 
 IDENT_CHAR = Regex("[@_$0-9A-Za-zÀ-ÖØ-öø-ƿ]").expr.parser_config.include
+FIRST_IDENT_CHAR = "".join(set(IDENT_CHAR) - set("0123456789"))
 SQL_NULL = Call("null", [], {})
 
 null_locations = []
@@ -37,11 +37,22 @@ null_locations = []
 def keyword(keywords):
     return And([
         Keyword(k, caseless=True) for k in keywords.split(" ")
-    ]).set_parser_name(keywords)
+    ]).set_parser_name(keywords) / (lambda: keywords.replace(" ", "_"))
+
+
+def flag(keywords):
+    """
+    RETURN {keywords: True}
+    """
+    return (keyword(keywords) / (lambda: True))(keywords.replace(" ", "_"))
+
+
+def assign(key: str, value: ParserElement):
+    return keyword(key).suppress() + value(key.replace(" ", "_"))
 
 
 def simple_op(op, args, kwargs):
-    if not args:
+    if args is None:
         kwargs[op] = {}
     else:
         kwargs[op] = args
@@ -118,6 +129,11 @@ def _chunk(values, size):
         yield acc
 
 
+def to_lambda(tokens):
+    params, op, expr = list(tokens)
+    return Call("lambda", [expr], {"params": list(params)})
+
+
 def to_json_operator(tokens):
     # ARRANGE INTO {op: params} FORMAT
     length = len(tokens.tokens)
@@ -170,6 +186,8 @@ def to_json_operator(tokens):
             while isinstance(operand, ParseResults) and isinstance(operand.type, Group):
                 # PARENTHESES CAUSE EXTRA GROUP LAYERS
                 operand = operand[0]
+                if isinstance(operand, ParseResults) and isinstance(operand.type, Forward):
+                    operand = operand[0]
 
             if isinstance(operand, Call) and operand.op == op:
                 acc.extend(operand.args)
@@ -183,6 +201,16 @@ def to_json_operator(tokens):
     return binary_op
 
 
+def to_offset(tokens):
+    expr, offset = tokens.tokens
+    return Call("get", [expr, offset], {})
+
+
+def to_window_mod(tokens):
+    expr, window = tokens.tokens
+    return Call("value", [expr], {**window})
+
+
 def to_tuple_call(tokens):
     # IS THIS ONE VALUE IN (), OR MANY?
     tokens = list(tokens)
@@ -234,6 +262,7 @@ binary_ops = {
     "not_simlilar_to": "not_similar_to",
     "or": "or",
     "and": "and",
+    "->": "lambda",
     "union": "union",
     "union_all": "union_all",
     "union all": "union_all",
@@ -256,18 +285,19 @@ def to_json_call(tokens):
     # ARRANGE INTO {op: params} FORMAT
     op = tokens["op"].lower()
     op = binary_ops.get(op, op)
-
-    params = listwrap(tokens["params"])
-    if tokens["ignore_nulls"]:
-        ignore_nulls = True
+    params = tokens["params"]
+    if isinstance(params, (dict, str, int, Call)):
+        args=[params]
     else:
-        ignore_nulls = None
+        args = list(params)
+
+    kwargs = {k: v for k, v in tokens.items() if k not in ("op", "params")}
 
     return ParseResults(
         tokens.type,
         tokens.start,
         tokens.end,
-        [Call(op, params, {"ignore_nulls": ignore_nulls})],
+        [Call(op, args, kwargs)],
         tokens.failures,
     )
 
@@ -324,24 +354,26 @@ def to_join_call(tokens):
 
 
 def to_expression_call(tokens):
-    over = tokens["over"]
-    within = tokens["within"]
-    if over or within:
+    if set(tokens.keys()) & {"over", "within", "filter"}:
         return
 
-    expr = ParseResults(
+    return ParseResults(
         tokens.type,
         tokens.start,
         tokens.end,
         listwrap(tokens["value"]),
         tokens.failures,
     )
-    return expr
+
+
+def to_over(tokens):
+    if not tokens:
+        return {}
 
 
 def to_alias(tokens):
     cols = tokens["col"]
-    name = tokens[0][0]
+    name = tokens["name"]
     if cols:
         return {name: cols}
     return name
@@ -351,7 +383,9 @@ def to_top_clause(tokens):
     value = tokens["value"]
     if not value:
         return None
-    elif tokens["ties"]:
+
+    value = value.value()
+    if tokens["ties"]:
         output = {}
         output["ties"] = True
         if tokens["percent"]:
@@ -365,24 +399,70 @@ def to_top_clause(tokens):
         return [value]
 
 
-def to_select_call(tokens):
-    value = tokens["value"]
-    if value.value() == "*":
-        return ["*"]
-
-    if value["over"] or value["within"]:
-        output = ParseResults(
-            tokens.type, tokens.start, tokens.end, value.tokens, tokens.failures
-        )
-        output["name"] = tokens["name"]
-        return output
+def to_row(tokens):
+    columns = list(tokens)
+    if len(columns) > 1:
+        return {"select": [{"value": v[0]} for v in columns]}
     else:
-        return
+        return {"select": {"value": columns[0]}}
+
+
+def to_values(tokens):
+    rows = list(tokens)
+    if len(rows) > 1:
+        return {"union_all": list(tokens)}
+    else:
+        return rows
+
+
+def to_stack(tokens):
+    width = tokens["width"]
+    args = listwrap(tokens["args"])
+    return Call("stack", args, {"width": width})
+
+
+def to_array(tokens):
+    types = list(tokens["type"])
+    args = list(tokens["args"])
+    output = Call("create_array", args, {})
+    if types:
+        output = Call("cast", [output, Call("array", types, {})], {})
+    return output
+
+
+def to_map(tokens):
+    keys = tokens["keys"]
+    values = tokens["values"]
+    return Call("create_map", [keys, values], {})
+
+
+def to_struct(tokens):
+    types = list(tokens["types"])
+    args = list(d for a in tokens["args"] for d in [a if a['name'] else a['value']])
+
+    output = Call("create_struct", args, {})
+    if types:
+        output = Call("cast", [output, Call("struct", types, {})], {})
+    return output
+
+
+def to_select_call(tokens):
+    expr = tokens['value']
+    if expr == "*":
+        return ["*"]
+    try:
+        call = expr[0][0]
+        if call.op == "value":
+            return {"name": tokens['name'], "value": call.args, **call.kwargs}
+    except:
+        pass
 
 
 def to_union_call(tokens):
     unions = tokens["union"]
-    if unions.type.parser_name == "unordered sql":
+    if isinstance(unions, dict):
+        return unions
+    elif unions.type.parser_name == "unordered sql":
         output = {k: v for k, v in unions.items()}  # REMOVE THE Group()
     else:
         unions = list(unions)
@@ -408,10 +488,23 @@ def to_union_call(tokens):
     return output
 
 
-def to_statement(tokens):
+def to_query(tokens):
     output = tokens["query"][0]
-    output["with"] = tokens["with"]
-    return output
+    try:
+        output["with"] = tokens["with"]
+        output["with_recursive"] = tokens["with_recursive"]
+
+        return output
+    except Exception as cause:
+        return
+
+
+def to_table(tokens):
+    output = dict(tokens)
+    if len(list(output.keys())) > 1:
+        return output
+    else:
+        return output["value"]
 
 
 def unquote(tokens):
